// This toy example shows how to define an ostream-like class in such
// a way that everything that will be written to the stream goes both
// to the standard output and to a disk log file.

#include <iostream>
#include <fstream>
#include <stdexcept>

// The important part of this file is a class derived from the class
// streambuf defined in <iostream>; streambuf defines an interface
// used by [i/o]streams to read bytes from an external representation
// and to write bytes to an external representation.  The actual
// handling of the external representation is implemented in classes
// derived from streambuf; in our case the "external representation"
// is a wrapper class that will send every received byte to two
// streams (only _output_ streams are considered).  An internal flag
// allows to inhibit output to the log file; methods to access and
// modify that flag are provided.
//
// The hard part: streambuf and ostream are template classes defined
// as a basic_streambuf of chars and a basic_stream of chars; we have
// to start from the beginning.  Here is a template class that
// inherits from basic_streambuf and that has as members two more
// basic_streambuf's ("out1" corresponding to the "screen", and "out2"
// corresponding to the "log file"); the constructor must be called
// with two valid basic_streambuf's.

template < class C, class T = std::char_traits< C > >
class basic_teebuf : public std::basic_streambuf< C, T >
{
private:
  typedef std::basic_streambuf< C, T >      buf_type;
  typedef typename buf_type::char_type      char_type;
  typedef typename buf_type::int_type       int_type;
  typedef typename buf_type::traits_type    traits_type;

  int_type       overflow(int_type c = traits_type::eof());
  int_type       underflow();
  int_type       uflow();

  buf_type      *out1;
  buf_type      *out2;
  bool           tee;

public:
  basic_teebuf(buf_type *a, buf_type *b)
    : out1(a), out2(b), tee(true) {}

  void closeTee()        { tee = false; }
  void openTee()         { tee = true;  }
  bool teeOpened() const { return tee;  }
};

// The basic_teebuf method that, called for every character to be put
// in the streambuf, will put it on both the "screen" and the "log
// file" streambuf

template < class C, class T >
typename basic_teebuf< C, T >::int_type
basic_teebuf< C, T >::overflow(
  typename basic_teebuf< C, T >::int_type c
) {
  int_type retVal( traits_type::not_eof(c) );

  if (       out1->sputc(c) == traits_type::eof()) retVal = traits_type::eof();
  if (tee && out2->sputc(c) == traits_type::eof()) retVal = traits_type::eof();
  return retVal;
}

// The methods called for getting a character from the streambuf: they
// both throw an exception (we don't want to read from the stream)

template < class C, class T >
typename basic_teebuf< C, T >::int_type
basic_teebuf< C, T >::underflow() {
  throw std::runtime_error("Can't happen (underflow)");
}

template < class C, class T >
typename basic_teebuf< C, T >::int_type
basic_teebuf< C, T >::uflow() {
  throw std::runtime_error("Can't happen (uflow)");
}

// The ostream-like class that wraps the basic_teebuf

template < class C, class T = std::char_traits< C > >
class basic_teestream : public std::basic_ostream< C, T >
{
private:
  basic_teebuf< C, T >    buf;

public:
  basic_teestream(std::basic_ostream< C, T > &a,
                  std::basic_ostream< C, T > &b)
    : std::basic_ostream< C, T >(&buf), buf(a.rdbuf(), b.rdbuf()) {}

  void closeTee()        { buf.closeTee();       }
  void openTee()         { buf.openTee();        }
  bool teeOpened() const { return buf.teeOpened; }
};

typedef basic_teebuf< char >      teebuf;
typedef basic_teestream< char >   teestream;

// The test program.  Opens a file called "logfile.txt" and outputs
// three lines of shit to the screen (two of them go to the log file
// too).

#include <ctime>

int main() {
  const time_t now( time(0) );
  const unsigned int weekDays = 7;
  std::ofstream log("logfile.txt");
  if (log) {
    teestream lo(std::cout, log);
    lo << "Hi, MLO;";
    lo << ' ' << "how are you ?" << std::endl;
    lo.closeTee();
    lo << "In a week there are " << weekDays << " days\n";
    lo.openTee();
    lo << "Now is " << asctime(localtime(&now));
    return 0;
  } else {
    std::cerr << "Couldn't open log file\n";
    return EXIT_FAILURE;
  }
}
