// First all needed header files are included: The file prfxstream.hh declares
// the classes whose member functions are implemented here. In addition to this
// some functions from std::string are use to handle strings.

#include <string>
#include "prfxstream.hh"

//----------------------------------------------------------------------------
// The constructor of the prfxbuf initializes its pointer to the streambuf
// with the argument sb: It is assumed that this streambuf is initialized
// correctly.  In addition no ownership is assumed for this streambuf. It
// is not deleted in the destructor.  Then the length of the prefix string
// is cached and the prefix string is copied into a private version: This
// is done to avoid problems when the user modifies or deletes the string
// passed as constructor argument. The member i_newline is set to indicate
// that the processing it at the beginning of a new line: in either case
// (reading or writing) it starts with a new line. When reading a file a
// prefix has to be skipped and when writing a file a prefix has to be
// added. EOF is used to indicate that the cache does not contain any
// valid character.
// In the body of the constructor the put area and the get area are
// initialized to be empty: no buffering is done by this streambuf. All
// buffering is deferred to the actually used streambuf. This makes sure
// that the function overflow() is called whenever a character is written
// to this streambuf and that the function underflow() is called whenever
// a character is read from this streambuf. The put buffer is specified
// using streambuf::setp() and the get buffer is specified using
// streambuf::setg().

prfxbuf::prfxbuf(
  streambuf *sb,
  const char *prfx
) : streambuf(),
    i_sbuf(sb),
    i_len(strlen(prfx)),
    i_prfx(strcpy(new char[i_len + 1], prfx)),
    i_buffer(new char[i_len]),
    i_newline(true),
    i_cache(EOF)
{
  setp(0, 0);
  setg(0, 0, 0);
}

// The destructor of prfxbuf has to release the copy of the prefix.

prfxbuf::~prfxbuf()
{
  delete[] i_prfx;
  delete[] i_buffer;
}

// ----------------------------------------------------------------------------
// The function skip_prefix() tries to skip a prefix at the beginning of the
// pending input. To do so it attempts to read as many characters from the
// associated streambuf as the prefix has. If this succeeds the read characters
// are compared with the expected prefix. If either end of file is encountered
// while reading characters or the read characters don't match the prefix false
// is returned: this indicates end of file for the prefix stream (EOF is
// handled by the two function uflow() and underflow() which call
// skip_prefix()).

bool prfxbuf::skip_prefix()
{
  if (i_sbuf->sgetn(i_buffer, i_len) != i_len) return false;
  if (strncmp(i_buffer, i_prfx, i_len)) {
    // an expection could be thrown here...
    return false;
  } else {
    i_newline = false;
    return true;
  }
}

// ----------------------------------------------------------------------------
// The function underflow() makes the next character available if no characters
// is found in the get area. Since the get area is always empty for a prfxbuf
// (see the constructor) this function is called when the current character is
// accessed without moving the stream to the next character (see uflow() below
// for a slightly modified behavior). The function returns the character at the
// current position or EOF if there is no character at this position left.
// The function checks whether a character is currently cached. If this is the
// case this character is simply returned. Otherwise a prefix at the beginning
// of a line is skipped if this is necessary. Then one character is extracted
// from the used streambuf using streambuf::sbumpc(). This character is stored
// in i_cache to have this character available if underflow() is called for the
// same position again. Then it is checked whether it is the end of a line and
// i_newline is set to true if it actually is the end of a line. Finally, the
// cached character is returned.
// If it wasn't possible to skip the prefix or if there is no character
// available from the used streambuf, EOF is returned. This indicates that end
// of file has been reached.

int prfxbuf::underflow()
{
  if (i_cache == EOF) {
    if (i_newline)
      if (!skip_prefix())
        return EOF;

    i_cache = i_sbuf->sbumpc();
    if (i_cache == '\n')
      i_newline = true;
    return i_cache;
  } else {
    return i_cache;
  }
}

// ----------------------------------------------------------------------------
// The behavior of uflow() is similar to the behavior of underflow(). The only
// difference between these two functions is that uflow() reads past the
// current character. This means that a call to uflow() changes the current
// character to be the next character in the stream.
// The actual implementation of this function is similar to that of
// underflow(): if a character is left in the cache (by underflow()) uflow()
// clears the cache and returns the character stored in the cache. Otherwise
// the prefix is skipped (if necessary) and the next character is extracted
// from the used streambuf using streambuf::sbumpc(). It is checked whether the
// extracted character indicates end of line and if so i_newline is set to
// true. Finally the result is returned: the character returned from sbumpc()
// or EOF (if skipping the prefix failed). The value returned from sbumpc() may
// indicate EOF for the used streambuf.

int prfxbuf::uflow()
{
  if (i_cache == EOF) {
    if (i_newline)
      if (!skip_prefix())
        return EOF;

    int rc = i_sbuf->sbumpc();
    if (rc == '\n')
      i_newline = true;
    return rc;
  }

  int rc = i_cache;
  i_cache = EOF;
  return rc;
}

// ----------------------------------------------------------------------------
// The function overflow() is concerned with overflow of the put buffer: This
// function is called with the next character to put into the stream if no more
// positions are available in the put area. Since the put area is empty for
// prfxbufs this function is always called if another character is to be put
// into the stream.
// The operation of overflow() is simple: if some other character than EOF is
// to be put into the stream, this character is put into the used stream using
// streambuf::sputc(). If there was a pending newline, the character is
// preceeded by the prefix which is put into the stream using
// streambuf::sputn(). After putting the character into the stream it is
// checked whether this character is an end of line: in this case i_newline is
// set to true to indicate that a newline is pending. The prefix is not written
// immediately to avoid writing a prefix on the last line in a file.
// If an error occures, either when putting the prefix or the next character
// into the stream, EOF is returned.

int prfxbuf::overflow(int c)
{
  if (c != EOF) {
    if (i_newline)
      if (i_sbuf->sputn(i_prfx, i_len) != i_len)
        return EOF;
      else
        i_newline = false;

    int rc = i_sbuf->sputc(c);
    if (c == '\n')
      i_newline = true;
    return rc;
  }
  return 0;
}

// ----------------------------------------------------------------------------
// sync() is used to bring the internal representation into synchronization
// with the external representation. Since a prfxbuf has no internal
// representation other than the used streambuf all which has to be done is to
// call sync() for the used streambuf.

int prfxbuf::sync()
{
  i_sbuf->sync();
  return 0;
}

// ----------------------------------------------------------------------------
// The only missing functions are the constructors and the destructors of the
// classes iprfxstream (used to read a file stripping of a prefix from every
// line read) and oprfxstream (used to prepend a prefix to every line written).
// The constructors are doing exactly the same thing: They initialize their
// base class with a prfxbuf which is constructed from their arguments.

iprfxstream::iprfxstream(
  streambuf *sb,
  const char *prfx
) : istream(new prfxbuf(sb, prfx))
{ }

oprfxstream::oprfxstream(
  streambuf *sb,
  const char *prfx
) : ostream(new prfxbuf(sb, prfx))
{ }

// ----------------------------------------------------------------------------
// Finally, the destructors of the class iprfxstream and oprfxstream delete the
// prfxbuf which was constructed in the constructor. The pointer to the prfxbuf
// is obtained using the ios::rdbuf(): The actual pointer is stored in the ios
// part of the [io]prfxstream. There is no need to duplicate this pointer.

iprfxstream::~iprfxstream()
{
  delete rdbuf();
}

oprfxstream::~oprfxstream()
{
  delete rdbuf();
}
