// This page includes a header file used to declare a specialized istream,
// ostream, and streambuf: The prfxbuf derived from streambuf is used to put a
// prefix before any line written and to strip off a prefix from any line read.
// The [io]streams are a convenient interface to this prfxbuf. This file only
// contains the declarations. The definitions of the functions are found in the
// file prfxstream.cc.
//
// For further information about iostreams see a documentation of this stuff.
// Most of the knowledge I have about iostreams comes from
// C++ IOStreams Handbook
// Steve Teale
// Addison-Wesley
// ISBN 0-201-59641-5
// ----------------------------------------------------------------------------
// First prevent that the body of this file is included twice:

#ifndef PRFXSTREAM_HH_
#define PRFXSTREAM_HH_

// Next the necessary header file is included: iostream is needed for the
// declaration of ostream and streambuf.

#include <iostream>

using namespace std;

// ----------------------------------------------------------------------------
// The important part of this file is a class derived from the class streambuf:
// The class streambuf defines an interface used by [io]streams to read bytes
// from an external representation and to write bytes to an external
// representation. The actual handling of the external representation is
// implemented in classes derived from streambuf. In this case the "external
// representation" is a wrapper for another external representation used to
// prepend a prefix to every line written and to strip off a prefix from every
// line read. Another streambuf is used by this class to handle a real external
// representation.

class prfxbuf: public streambuf {
private:
  streambuf    *i_sbuf;         // the actual streambuf used to read and
                                // write chars
  int           i_len;          // the length of the prefix
  char         *i_prfx;         // the prefix
  char         *i_buffer;       // the prefix buffer
  bool          i_newline;      // remember whether we are at a new line
  int           i_cache;        // may cache a read character

  bool          skip_prefix();

protected:
  int           overflow(int);
  int           underflow();
  int           uflow();
  int           sync();

public:
  prfxbuf(streambuf *sb, const char *prfx);
  ~prfxbuf();
};

// ----------------------------------------------------------------------------
// To provide convenient access to the prfxbuf there is an istream and an
// ostream provided: both create a prfxbuf with the given arguments and
// initialize their base class (the istream or the ostream) with this object.
// Both class also take care that the allocate prfxbuf is destructed when the
// [io]stream is destructed. It is also possible to perform the same operations
// "by hand" i.e. to construct a prfxbuf, initialize a corresponding stream
// with it, and finally, after the stream is no longer needed, destroy the
// created the constructed prfxbuf. However, it is more convenient and safer to
// use classes derived from [io]stream.

class iprfxstream: public istream {
public:
  iprfxstream(streambuf *sb, const char *prfx);
  ~iprfxstream();
};

// ----------------------------------------------------------------------------

class oprfxstream: public ostream {
public:
  oprfxstream(streambuf *sb, const char *prfx);
  ~oprfxstream();
};

// ----------------------------------------------------------------------------
// Finally there has to be a proper termination of the header file:

#endif // PRFXSTREAM_HH_
